"""
Gondola I/O system. Read files and connect to sockets to 
obtain various flavors of data used trhoughout GAPS.
"""

from .. import _gondola_core  as _gc 

import sys
if sys.version_info.minor <= 10:
    from datetime import datetime, timezone
    UTC = timezone.utc
else:
    from datetime import datetime, UTC, timezone

from pathlib import Path
import numpy as np

# shortcut for import
get_all_telemetry_event_names = _gc.io.get_all_telemetry_event_names
#read_example                  = _gc.io.read_example
get_runfilename               = _gc.io.get_runfilename 
get_califilename              = _gc.io.get_califilename
CRFrameObject                 = _gc.io.CRFrameObject 
DataSourceKind                = _gc.io.DataSourceKind 
CRReader                      = _gc.io.CRReader
CRWriter                      = _gc.io.CRWriter
CRFrame                       = _gc.io.CRFrame
TofPacketReader               = _gc.io.TofPacketReader
TelemetryPacketReader         = _gc.io.TelemetryPacketReader
list_path_contents_sorted     = _gc.io.list_path_contents_sorted
get_utc_now                   = _gc.io.get_utc_timestamp 
get_utc_date                  = _gc.io.get_utc_date
get_datetime                  = _gc.io.get_datetime
get_rundata_from_file         = _gc.io.get_rundata_from_file
get_unix_timestamp            = _gc.io.get_unix_timestamp 

def get_ts_from_binfile(fname):
    """ 
    Get a timestamp from a '.bin' file as written 
    by telemetry

    # Arguments:
        * fname  : name of the file 
    
    # Returns:
        datetime.datetime (UTC) 
    """
    ts = str(fname)[-17:-4]
    ts = get_unix_timestamp(ts)
    return datetime.fromtimestamp(ts, UTC)

#---------------------------------------------------

def get_telemetry_binaries(unix_time_start, unix_time_stop,\
                           data_dir='/gaps_binaries/live/raw/ethernet'):
    """
    Get the relevant telemetry data files for time period from a directory

    # Arguments
        * unix_time_start : seconds since epoch for run start
        * unix_time_end   : seconds since epoch for run end

    # Keyword Arguments
        * data_dir        : folder with telemetry binaries ('.bin')
    """
    # file format is something like RAW240712_094325.bin
    t_start = datetime.fromtimestamp(unix_time_start, UTC)
    t_stop  = datetime.fromtimestamp(unix_time_stop, UTC)
    all_files = sorted([k for k in Path(f'{data_dir}').glob('*.bin')])
    #print(f'-> Found {len(all_files)} files in {data_dir}')
    # heres some new logic yikes
    ts = np.array([get_ts_from_binfile(f) for f in all_files])
    all_files = np.array(all_files)[np.argsort(ts)]
    ts = np.sort(ts)
    i = 0
    while(ts[i]<t_start):
        i+=1
    j=i
    while(ts[j]<t_stop):
        j+=1
    i-=1
    j+=1
    # FiXME - this might throw away 1 file
    #files = [f for f, ts in zip(all_files, ts) if t_start <= ts <= t_stop]
    files = all_files[i:j]
    ts = [get_ts_from_binfile(f) for f in files]
    print(f'-> Run duration {ts[-1] - ts[0]}')
    if len(files)>0:
        print(f'-> Found {len(files)} files within range of {t_start} - {t_stop}')
        print(f'--> Earliest file {files[0]}')
        print(f'--> Latest file {files[-1]}')
    else:
        print(f'! No files have been found within {t_start} and {t_stop}!')
    return files

def grace_get_telemetry_binaries(unix_time_start, unix_time_stop,\
                                 data_dir='/gaps_binaries/live/raw/ethernet'):

    # file format is something like RAW240712_094325.bin
    t_start = datetime.fromtimestamp(unix_time_start, UTC)
    t_stop  = datetime.fromtimestamp(unix_time_stop, UTC)
    all_files = sorted([k for k in Path(f'{data_dir}').glob('*.bin')])
    print(f'-> Found {len(all_files)} files in {data_dir}')
    #ts = [get_ts_from_binfile(f) for f in all_files]
    # FiXME - this might throw away 1 file
    #files = [f for f, ts in zip(all_files, ts) if t_start <= ts <= t_stop]
    #ts = [get_ts_from_binfile(f) for f in files]
    time_stamps = [get_ts_from_binfile(f) for f in all_files]
    # I think there is probably a more elegant way to do this with numpy sorting
    # but dont know how it handles datetime object, ona  side note
    # why tf are we using datetime objects instead of like floats...
    sorted_inds = np.argsort(time_stamps)
    sorted_files = np.array(all_files)[sorted_inds]
    sorted_times = np.array(time_stamps)[sorted_inds]
    i = np.argmin(np.abs(t_start-sorted_times))-1
    j = np.argmin(np.abs(t_stop-sorted_times))+1
    if i < 1:
        i = 1
    if j >= len(sorted_files):
        j = len(sorted_files)-1
    files = sorted_files[i-1:j+1]
    print(f'-> Run duration {sorted_times[j] - sorted_times[i-1]}')
    if len(files) > 0:
        print(f'-> Found {len(files)} files within range of {t_start} - {t_stop}')
        print(f'--> Earliest file {files[0]}')
        print(f'--> Latest file {files[-1]}')
    else:
        print(f'! No files have been found within {t_start} and {t_stop}!')
    return files





"""
gondola - GAPS online software core suite for python.
"""

# GAPS online/offline software core suite for python

from . import gondola_core as _gondola_core

import importlib as _importlib
import os as _os 

__version__ = _gondola_core.get_version()

events     = _gondola_core.events 
packets    = _gondola_core.packets 
#io         = _gondola_core.io 
monitoring = _gondola_core.monitoring 
stats      = _gondola_core.stats 
algo       = _gondola_core.algo 
#db         = _gondola_core.db 
#tracker    = _gondola_core.tracker 

# set up the python submodules
from . import calibration 
from . import visual
from . import io
from . import tof
from . import db 
from . import reconstruction 
from . import tracker

#calibration.RBCalibrations = _gondola_core.calibration.RBCalibrations

__all__ = ['events', 'packets', 'io', 'monitoring', 'stats', 'algo', 'db',
           'calibration', 'visual']

# clean up the namespace, module still available as hidden through _gondola_core
del gondola_core

def init_database():
    """
    Returns the path to the included SQLite database file.
    """
    with _importlib.resources.path("gondola", "gaps_flight.db") as db_path:
        _os.environ['GONDOLA_DB_URL'] = str(db_path) 
        return db_path

#----------------------------------
# Initializing 

init_database() 

print (f'Welcome to gondola v{__version__}, a software suite for the \U0001F388 GAPS experiment! Bulld for \U0001F40D with the power of \U0001F980! \u2728')
print (f' -- The database has been set to GONDOLA_DB_URL {_os.environ["GONDOLA_DB_URL"]}')




"""
Tracker related plots and visualizations
"""

from .. import _gondola_core as _gc 

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import charmingbeauty as cb 
import charmingbeauty.layout as lo
from matplotlib import patches

from loguru import logger
from copy import deepcopy as copy

strip_lines = _gc.tracker.strip_lines

def plot_strip_lines(ax, det, layer : int, color='k'):
    """
    Use lines to indicate detector strips. Automatically apply 
    correct orientation for even/odd layers.

    # Arguments:
        ax    : axis instance to plot on 
        det   : iterable providing x,y coordinates
        layer : tracker layer (0-9)

    # Keyword Arguments:
        color : the color of the lines
    """
    strip_widths = strip_lines()
    if layer % 2 == 0: # even layer
        for r,sw in strip_widths:
            ax.plot([det[0] + sw, det[0] + sw], [det[1] - r, det[1] + r], color=color, alpha=0.4, lw=0.5)
    else:
        for r,sw in strip_widths:
            ax.plot([det[0] - r, det[0]+ r], [det[1] + sw, det[1] + sw], color=color, alpha=0.4, lw=0.5)

#-------------------------------------------------------

def plot_tracker_proj(ax,\
                      hits,\
                      projection   = "xy",\
                      use_energy   = False,\
                      cmap         = matplotlib.colormaps['seismic'],\
                      circle_color ='w',\
                      color_energy = False,\
                      cnorm_max    = None,\
                      hitstyle     = {'edgecolor': 'w', 'alpha' : 0.5 , 'marker' : 'o'} ):
    """
    Plot 2d projection of the tracker on the given axis. Hits can be indicated if 
    available. Use together with "prepare_layer_fig" for a full plot with 
    labeled axes.

    # Arguments: 
      * ax         : The axes to plot on 
      * hits       : An iterable of gondola_core.events.TrackerHit

    # Keyword Arguments:
      * projection   : The projection to plot. Can be either "xy", "xz" or "yz" 
      * cmap         : A matplotlib compatible colormap to color the hits
      * circle_color : The color with the circle showing the detector outlines 
                       are indicated 
      
      * color_energy : Color the circles for the hits with the energy deposition
                       FIXME - color scale needs to be normalized together with the 
                       TOF
      * cnorm_max    : The colors can be normalized. Specify a maximum for the normalization
                       (min will be 0)
      * hitstyle     : Specify the style of the plotted hits by a dictionary which will be 
                       passed on to plt.scatter
    """
    if not projection in ["xy", "xz", "yz"]:
        raise ValueError("projection argument needs to be either 'xy', 'xy' or 'xz'!")

    hit_layers = list(set([k.layer for k in hits]))
    strip_dict = _gc.db.TrackerStrip.all_as_dict() 
    det_coords = [(k.global_pos_x_det_l0, k.global_pos_y_det_l0, k.global_pos_z_det_l0) for k in strip_dict.values()]
    det_coords = [k for k in set(det_coords)]
    strip_coord     = [strip_dict[k].coordinates for k in strip_dict]
    hit_strip_coord = [strip_dict[h.strip_id].coordinates for h in hits]
    hit_det_coord   = [strip_dict[h.strip_id].detector_coordinates for h in hits]
    #def get_stripcoordinates(h : _gc.events.TrackerHit):
    #    geo = strip_dict[h.strip_id] 
    #    return (geo.global_pos_x_l0, geo.global_pos_y_l0, geo.global_pos_z_l0)
    #
    #def get_detcoordinates(h : _gc.events.TrackerHit):
    #    geo = strip_dict[h.strip_id]
    #    return (geo.global_pos_x_det_l0, geo.global_pos_y_det_l0, geo.global_pos_z_det_l0)

    #all_dets = [(k.global_pos_x_det_l0, k.global_pos_y_det_l0, k.global_pos_z_det_l0) for k in go.db.TrackerStrip.objects.all()]
    idx = 0,1 
    match projection:
        case "xy":
            pass
        case "xz":
            idx = 0,2
        case "yz":
            idx = 1,2 
    # this is for all detectors   
    all_dets_coord = set([(k[idx[0]],k[idx[1]]) for k in det_coords])
   
    # this is for hit detectors
    #strip_coord = [get_stripcoordinates(k) for k in hits]
    #det_coord   = [get_detcoordinates(k) for k in hits]
    adc         = [k.adc for k in hits]
    if use_energy: # this requires that the hit underwent coordinates
        adc = [k.energy for k in hits]
    else:
        adc = np.array(adc) / 12
    xs = [h.x for h in hits]
    ys = [h.y for h in hits]
    zs = [h.z for h in hits]
    if color_energy:
        if cnorm_max != None:
            cm_norm_pts = plt.Normalize(vmin=0, vmax=cnorm_max)
        else:
            cm_norm_pts = plt.Normalize(vmin=min(adc), vmax=max(adc))
    # construct the detector plot
    match projection:
        case "xy":
            for k in all_dets_coord:
                patch = patches.Circle(k, radius=5, fill=False, color='gray', alpha=0.1)
                for layer in hit_layers:
                    plot_strip_lines(ax, k, layer, color='gray')
                ax.add_patch(patch)
            # only for hit strips
            for k in hit_det_coord:
                patch = patches.Circle(k, radius=5, fill=False, color=circle_color)
                ax.add_patch(patch)
                for layer in hit_layers:
                    plot_strip_lines(ax, k, layer, color='gray')
            if not color_energy:
                ax.scatter(xs, ys, marker=hitstyle['marker'], s=adc, facecolor='none', edgecolor=hitstyle['edgecolor'])
            else:
                ax.scatter(xs, ys, marker=hitstyle['marker'], s=100*adc,\
                           edgecolor=hitstyle['edgecolor'], color=cmap(cm_norm_pts(adc)))
        case "xz":
            for k in all_dets_coord:
                det = [k[0] - 5, k[1]]
                rect_patch = patches.Rectangle(det, width=10, height=0.25, color='gray', alpha=0.5)
                ax.add_patch(rect_patch)
            # only for hits 
            for k in hit_det_coord:
                det = [k[0] - 5, k[2]]
                rect_patch = patches.Rectangle(det, width=10, height=0.25, color=circle_color, alpha=0.8)
                ax.add_patch(rect_patch)
            if not color_energy:
                ax.scatter(xs, zs, marker=hitstyle['marker'], s=adc, facecolor='none', edgecolor=hitstyle['edgecolor'])
            else:
                ax.scatter(xs, zs, marker=hitstyle['marker'], s=100*adc,\
                           edgecolor=hitstyle['edgecolor'], color=cmap(cm_norm_pts(adc)))
        case "yz":
            for k in all_dets_coord:
                det = [k[0] - 5, k[1]]
                rect_patch = patches.Rectangle(det, width=10, height=0.25, color='gray', alpha=0.5)
                ax.add_patch(rect_patch)
            # only the hits 
            for k in hit_det_coord:
                det = [k[1] - 5, k[2]]
                rect_patch = patches.Rectangle(det, width=10, height=0.25, color=circle_color, alpha=0.8)
                ax.add_patch(rect_patch)
            if not color_energy:
                ax.scatter(ys, zs, marker=hitstyle['marker'], s=adc, facecolor='none', edgecolor=hitstyle['edgecolor'])
            else:
                ax.scatter(ys, zs, marker=hitstyle['marker'], s=100*adc,\
                           edgecolor=hitstyle['edgecolor'], color=cmap(cm_norm_pts(adc)))
    return ax

#-------------------------------------------------------

def prepare_layer_fig(layer, projection='XY'):
    """
    Set up figure and axis objects for a 2d projection
    of tracker layers.

    This will NOT plot any detectors/hits.
    It will ONLY prepare the figure/axes objects with 
    labels and such!

    # Keyword Arguments:
        * projection : Either 'XY', 'XZ' or 'ZY'. Show 
                       the tracker from the respective 
                       side
    """
    fig = plt.figure(figsize=lo.FIGSIZE_A4_SQUARE)
    ax  = fig.gca()
    ax.set_title(f'{layer}', loc='right')
    ax.spines['top'].set_visible(True)
    ax.spines['right'].set_visible(True)
    ax.grid(0)
    ax.set_aspect('equal')
    if projection == 'XY':
        ax.set_xlabel('X [cm]', loc='right')
        ax.set_ylabel('Y [cm]', loc='top')
        ax.set_xlim(-79,79)
        ax.set_ylim(-79,79) 
    if projection == 'XZ':
        ax.set_xlabel('X [cm]', loc='right')
        ax.set_ylabel('Z [cm]', loc='top')
    if projection == 'YZ':
        ax.set_xlabel('Y [cm]', loc='right')
        ax.set_ylabel('Z [cm]', loc='top')
    return fig, ax

#-------------------------------------------------------

def plot_tracker(hits = None,\
                 mask = None,\
                 circle_color='w') -> dict:
    """
    Create a number of 2d plots for the tracker. This includes
    individual plots for the individual layers as well as
    combined projections in xy, xz and yz. 
    
    # Keyword Arguments:
      * hits : A list of tracker hits. When given then 
               plot only those layers which have hits 
               If None given, plot all layers 
               (e.g. to be used with the mask argument)
      * mask : Indicate masked strips (grey them out) 

    """
    if not hits:
        logger.warning(f'No hits available for tracker layer plots!')
        return dict()
    strip_dict  = _gc.db.TrackerStrip.all_as_dict() 

    #n_layers   = len(set([k.layer for k in hits]))
    all_layers = list(set([k.layer for k in hits]))

    all_dets = []
    #all_strips = []

    all_dets = [(k.global_pos_x_det_l0, k.global_pos_y_det_l0, k.global_pos_z_det_l0) for k in strip_dict.values()]

    # this is used later for the individual layer plots
    #all_dets_xz = set([(k[0],k[2]) for k in all_dets])
    #all_dets_yz = set([(k[1],k[2]) for k in all_dets])
    all_dets_xy = set([(k[0],k[1]) for k in all_dets])

    # prepare output dictionary
    figures = dict()
 
    # prep the hits
    #strip_coord = [strip_dict[h.strip_id].coordinates for h in hits]
    #det_coord   = [strip_dict[h.strip_id].detector_coordinates for h in hits]
    #adc         = [k.adc for k in hits]
    #adc         = np.array(adc) / 12
    #xs          = [k[0] for k in strip_coord]
    #ys          = [k[1] for k in strip_coord]
    #zs          = [k[2] for k in strip_coord]
    
    # have one plot for the xy projection
    fig, ax     = prepare_layer_fig(layer='XY hit projection')
    plot_tracker_proj(ax, hits, projection='xy')
    figures['trk_proj_xy'] = copy(fig)
    
    # another plot for the xz projection
    fig, ax     = prepare_layer_fig(layer='XZ hit projection', projection='XZ')
    plot_tracker_proj(ax, hits, projection='xz')
    figures['trk_proj_xz'] = copy(fig)
    #ax.scatter(xs, zs, marker='o', s=adc, facecolor='none', edgecolor='r')
    ## strips
    #for k in all_dets_xz:
    #    det = [k[0] - 5, k[1]]
    #    rect_patch = patches.Rectangle(det, width=10, height=0.25, color='gray', alpha=0.1)
    #    ax.add_patch(rect_patch)
    #figures['trk_proj_xz'] = copy(fig)

    # another plot for the yz projection
    fig, ax     = prepare_layer_fig(layer='YZ hit projection', projection='YZ')
    plot_tracker_proj(ax, hits, projection='yz')
    figures['trk_proj_yz'] = copy(fig)
    #strip_coord = [get_stripcoordinates(k, strip_dict) for k in hits]
    #det_coord   = [get_detcoordinates(k, strip_dict) for k in hits]
    #adc = [k.adc for k in hits]
    #adc = np.array(adc) / 12
    #ys = [k[1] for k in strip_coord]
    #zs = [k[2] for k in strip_coord]
    #dets_ys = [k[0] for k in all_dets_yz]
    #dets_zs = [k[1] for k in all_dets_yz]
    # strips
    #for k in all_dets_yz:
    #    det = [k[0] - 5, k[1]]
    #    rect_patch = patches.Rectangle(det, width=10, height=0.25, color='gray', alpha=0.1)
    #    ax.add_patch(rect_patch)
    #ax.scatter(dets_ys, dets_zs, marker='+', color='gray')
    #ax.scatter(ys, zs, marker='o', s=adc, facecolor='none', edgecolor='r')
    #for k in det_coord:
    #    patch = patches.Circle(k, radius=5, fill=False, color='k')
    #    # im.set_clip_path(patch)
    #    ax.add_patch(patch)
    #figures['trk_proj_xz'] = copy(fig)

    # one plot per layer
    for __, layer in enumerate(all_layers):
        strip_coord = [strip_dict[h.strip_id].coordinates for h in hits if h.layer == layer]
        det_coord   = [strip_dict[h.strip_id].detector_coordinates for h in hits if h.layer == layer]
        adc         = [k.adc for k in hits if k.layer == layer]
        adc         = np.array(adc) / 12
        xs          = [k[0] for k in strip_coord]
        ys          = [k[1] for k in strip_coord]
        
        fig, ax = prepare_layer_fig(layer=f'Layer {layer}') 
        ax.scatter(xs, ys, marker='o', s=adc, facecolor='none', edgecolor='r')
        ax.scatter(xs, ys, marker='.', color='k', alpha=0.1)
        ax.set_title(f'Layer {layer}', loc='right')
        for k in det_coord:
            patch = patches.Circle(k, radius=5, fill=False, color=circle_color)
            # im.set_clip_path(patch)
            ax.add_patch(patch)
            #go.tracker.visual.plot_strip_lines(ax,k, layer, color=circle_color)
            plot_strip_lines(ax,k, layer, color=circle_color)
        for k in all_dets_xy:
            patch = patches.Circle(k, radius=5, fill=False, color='gray', alpha=0.1)
            # rect_patch = patches.Rectangle(k[1], width=1, height=100, color='gray', alpha=0.1)
            ax.add_patch(patch)
            # ax.add_patch(rect_patch)

        figures[f'trk_layer_{layer}'] = copy(fig)
    return figures

#-------------------------------------------------------

#import numpy as np
#import matplotlib.pyplot as plt
#import charmingbeauty.layout as lo 
#
#SILI_RADIUS = 5 # mm, with guardring and all
#
##-------------------------------------------------------
#

#-------------------------------------------------------




"""
Gondola core database functionality. This is a wrapper around 
the database functions in gondola-db
"""

from . import _gondola_core as _gc  
from matplotlib.patches import Rectangle

import vtk
import numpy as np


TofPaddle                    = _gc.db.TofPaddle
TofPaddle.__module__         = __name__
TofPaddle.__name__           = "TofPaddle"
TofPaddle.__doc__            = _gc.db.TofPaddle.__doc__
ReadoutBoard                 = _gc.db.ReadoutBoard
TrackerStrip                 = _gc.db.TrackerStrip
TrackerStripMask             = _gc.db.TrackerStripMask 
TrackerStripPedestal         = _gc.db.TrackerStripPedestal 
TrackerStripTransferFunction = _gc.db.TrackerStripTransferFunction 
TrackerStripCmnNoise         = _gc.db.TrackerStripCmnNoise
get_all_rbids_in_db          = _gc.db.get_all_rbids_in_db
get_hid_vid_map              = _gc.db.get_hid_vid_map
get_vid_hid_map              = _gc.db.get_vid_hid_map
get_dsi_j_ch_pid_map         = _gc.db.get_dsi_j_ch_pid_map

__all__ = ['TofPaddle']

#----------------------------------------
# extend the TofPaddles with a few methods
# to draw 3d/2d objects

def _create_box(self):

    # we can kind of cheat and do a cheap transform
    # the principal is either x, y or z axis, no mix-ins
    pr   = self.principal
    norm = self.normal
    cube = vtk.vtkCubeSource()
    edgepaddle = False
    if (pr == np.array([1,0,0])).all() or (pr == np.array([-1,0,0])).all():
        cube.SetXLength(self.length)  # Width in X
        if (norm == np.array([0,1,0])).all() or (norm == np.array([0,-1,0])).all():
            cube.SetYLength(self.height)
            cube.SetZLength(self.width)
        elif (norm == np.array([0,0,1])).all() or (norm == np.array([0,0,-1])).all():
            cube.SetZLength(self.height)
            cube.SetYLength(self.width)
        else:
            raise ValueError
    elif (pr == np.array([0,1,0])).all() or (pr == np.array([0,-1,0])).all():
        cube.SetYLength(self.length)  # Width in X
        if (norm == np.array([1,0,0])).all() or (norm == np.array([-1,0,0])).all():
            cube.SetZLength(self.width)
            cube.SetXLength(self.height)
        elif (norm == np.array([0,0,1])).all() or (norm == np.array([0,0,-1])).all():
            cube.SetXLength(self.width)
            cube.SetZLength(self.height)
        else:
            raise ValueError
    elif (pr == np.array([0,0,1])).all() or (pr == np.array([0,0,-1])).all():
        cube.SetZLength(self.length)
        # set the other two and then we have to rotat by 45 deg
        cube.SetXLength(self.height)
        cube.SetYLength(self.width)
        edgepaddle = True
        #transform_filter = vtk.vtkTransformPolyDataFilter()
        #transform_filter.SetInputData(cube.GetOutput())
        #transform_filter.SetTransform(trafo)
        #transform_filter.Update()
    else:
        ValueError(f'Unexpected principal axes {pr}')
    cube.Update()
    transform = vtk.vtkTransform()
    transform.Translate(self.global_pos_x_l0, self.global_pos_y_l0, self.global_pos_z_l0)
    if edgepaddle:
        transform.RotateWXYZ(45, [0,0,1])  # angle, (x, y, z) axis to rotate around
    transform_filter = vtk.vtkTransformPolyDataFilter()
    transform_filter.SetInputData(cube.GetOutput())
    transform_filter.SetTransform(transform)
    transform_filter.Update()
    #print (cube)
    # Extract the transformed points
    return transform_filter.GetOutput() 

TofPaddle._create_box = _create_box

def _create_box_points(self):
    points = self._create_box().GetPoints()
    #points = transform_filter.GetOutput().GetPoints()
    num_points = points.GetNumberOfPoints()
    transformed_points = np.array([points.GetPoint(i) for i in range(num_points)])
    return transformed_points
    #if vtk_import_success:
    #    # Use vtkCubeSource to create a cube (or box)
    #    cube.SetXLength(self.length)  # Width in X
    #    cube.SetYLength(self.width)  # Height in Y
    #    cube.SetZLength(self.height)  # Depth in Z
    #    cube.Update()         # Update the cube geometry
    #    return cube
    #else:
    #    raise NotImplementedError('This requires vtk to be available on your system!')

TofPaddle._create_box_points = _create_box_points
#TofPaddle.points      = None
#setattr(TofPaddle, "points", None)
#TofPaddle._points      = None
#def _set_points(self, points): 
#    self._points = points 
#TofPaddle.points = property(fget = lambda self : self._points, fset = _set_points, doc="points as for 2d/3d projection") 

#def _cache_box_points(self):
#    if self.points is None:
#        box    = self._create_box()
#        self.points = box
#
#TofPaddle._cache_box_points = _cache_box_points 

def _draw_xy(self, fill=False, lw=0.8, edgecolor='b', facecolor='b', alpha=0.7) -> Rectangle:
    """
    Draw a matplotlib patch for xy projection
    """
    #self._cache_box_points()
    box = self._create_box_points() 
    xy_points = box[:, :2]
    #if self.paddle_id in [57]:
    if (self.principal == np.array([0,0,1])).all() or (self.principal == np.array([0,0,-1])).all():
        xy_patch  = Rectangle(xy_points.min(axis=0), *(xy_points.max(axis=0) - xy_points.min(axis=0)),
                              fill=fill, edgecolor=edgecolor, facecolor=facecolor, lw=lw, alpha=alpha)
        # this is an edge paddle. For drawing use the angle feature of the Rectangle patch
        #xy_anchor0 = (self.global_pos_x_l0_A - self.width/np.sqrt(2), self.global_pos_y_l0_A - self.width/np.sqrt(2))
        corners = xy_patch.get_corners()
        # +x +y
        if self.paddle_id in [57,149,150,151]:
            anchor  = (corners[1][0], corners[1][1])
            rotation_point = (corners[1][0], corners[1][1])
            angle = 135
            xy_patch  = Rectangle(anchor, self.width, self.height,
                                  rotation_point = rotation_point, angle = angle,
                                  fill=fill, edgecolor=edgecolor, facecolor=facecolor, lw=lw, alpha=alpha)
        # -x -y
        if self.paddle_id in [58,152,153,154]:
            anchor  = (corners[0][0], corners[0][1])
            rotation_point = (corners[0][0], corners[0][1])
            angle = 45
            xy_patch  = Rectangle(anchor, self.width, self.height,
                                  rotation_point = rotation_point, angle = angle,
                                  fill=fill, edgecolor=edgecolor, facecolor=facecolor, lw=lw, alpha=alpha)
        # corner -x -y
        if self.paddle_id in [59,155,156,157]:
            anchor  = (corners[3][0], corners[3][1])
            rotation_point = (corners[3][0], corners[3][1])
            angle = -45
            xy_patch  = Rectangle(anchor, self.width, self.height,
                                  rotation_point = rotation_point, angle = angle,
                                  fill=fill, edgecolor=edgecolor, facecolor=facecolor, lw=lw, alpha=alpha)
        # corner +x -y
        if self.paddle_id in [60,158,159,160]:
            anchor  = (corners[3][0], corners[3][1])
            rotation_point = (corners[2][0], corners[2][1])
            angle = 45
            xy_patch  = Rectangle(anchor, self.width, self.height,
                                  rotation_point = rotation_point, angle = angle,
                                  fill=fill, edgecolor=edgecolor, facecolor=facecolor, lw=lw, alpha=alpha)

    else:
        xy_patch  = Rectangle(xy_points.min(axis=0), *(xy_points.max(axis=0) - xy_points.min(axis=0)),
                              fill=fill, edgecolor=edgecolor, facecolor=facecolor, lw=lw, alpha=alpha)
    return xy_patch

TofPaddle.draw_xy = _draw_xy


def _draw_xz(self, fill=False, lw=0.8, edgecolor='b', facecolor='b', alpha=0.7) -> Rectangle:
    """
    Draw a matplotlib patch for xy projection
    """
    box = self._create_box_points()
    xz_points = box[:, [0,2]]
    xz_patch  = Rectangle(xz_points.min(axis=0), *(xz_points.max(axis=0) - xz_points.min(axis=0)),
                          fill=fill, edgecolor=edgecolor, facecolor=facecolor, alpha=alpha, lw=lw)
    return xz_patch

TofPaddle.draw_xz = _draw_xz

def _draw_yz(self, fill=False, lw=0.8, edgecolor='b',facecolor='b', alpha=0.7) -> Rectangle:
    """
    Draw a matplotlib patch for xy projection
    """
    box = self._create_box_points()
    yz_points = box[:, 1:]
    yz_patch  = Rectangle(yz_points.min(axis=0), *(yz_points.max(axis=0) - yz_points.min(axis=0)),
                          fill=fill, edgecolor=edgecolor, facecolor=facecolor, alpha=alpha, lw=lw)
    return yz_patch

TofPaddle.draw_yz = _draw_yz




"""
Calibration related methods
"""

from pathlib import Path
import re
import tqdm

from . import _gondola_core 
RBCalibrations = _gondola_core.calibration.RBCalibrations

## convenience functions
def load_rb_calibrations(cali_dir : Path, load_event_data = False):
    """
    Load all calibrations stored in a certain directory and
    return a dictionary rbid -> RBCalibration

    # Arguments:
        * cali_dir        : Path with calibration files, one per RB

    # Keyword Arguments: 

        * load_event_data : if True, also load the associated events
                            which went into the calculation of the
                            calibration constants.
    """
    pattern = re.compile('RB(?P<rb_id>[0-9]*)_')
    calib_files = [k for k in cali_dir.glob("*.tof.gaps")]
    calibs = dict()
    for fname in tqdm.tqdm(calib_files, desc="Loading calibration files"):
        fname = str(fname)
        try:
            rb_id = int(pattern.search(fname).groupdict()['rb_id'])
        except Exception as e:
            print(f'Failed to get RB ID from file {fname}')   
            continue
        cali = RBCalibrations.from_file(fname)
        calibs[rb_id] = cali 
    return calibs




"""
TOF visualizations 
"""

from .. import _gondola_core as _gc 

import matplotlib
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import charmingbeauty as cb
import charmingbeauty.layout as lo
import numpy as np
import dashi
dashi.visual()

def plot_hg_lg_hits(h_nhits            : dashi.histogram.hist1d,
                    h_nthits           : dashi.histogram.hist1d,
                    n_events           : int,
                    no_hitmissing      = None,
                    one_hitmissing     = None,
                    lttwo_hitmissing   = None,
                    extra_hits         = None,
                    markercolor        = "w") -> (plt.Figure, plt.Figure):
    """
    Plot the HG vs the LG (trigger) hits with extra annotations.
    The nhit histograms and missing hit data has to be obtained
    previously.

    # Returns:
      Two figures, the actual histogram and the ratio plot 

    # Arguments:

      * h_nhits        : histogram of number of HG hits 
      * h_nthits       : histogram of number of LG hits
                         (hits with come from the trigger system
                            = 'Trigger hits')
      * n_events       : The number of events from which the 
                         hits were obtained
      * no_hitmissing  : The number of events which had zero
                         hits missing
      * one_hitmissing : The number of events which had one HG 
                         hit. missing 
      * two_hitmissing : The number of events which had two HG 
                         hits missing 
     """
    textbox = ''
    if n_events is not None:
        textbox  = f'NHits : {n_events:.2e}\n'
    if no_hitmissing is not None:
        textbox += f'{100*no_hitmissing/n_events:.2f}' + r'\%' + f' for N(LG) == N(HG)\n'
    if one_hitmissing is not None:
        textbox += f'{100*one_hitmissing/n_events:.2f}' + r'\%' + f' for N(LG) - N(HG) == 1\n'
    if lttwo_hitmissing is not None:
        textbox += f'{100*lttwo_hitmissing/n_events:.2f}' + r'\%' + f' for N(LG) - N(HG) $>=$ 2\n'
    if extra_hits is not None:
        textbox += f'{100*extra_hits/n_events:.2f}' + r'\%' + f' with N(HG) $>$ N(LG)\n'
    
    fig = plt.figure(figsize=lo.FIGSIZE_A4_LANDSCAPE)
    ax  = plt.gca()
    h_nhits .line(filled=True, alpha=0.7, color='tab:blue', label='HG')
    h_nthits.line(color='tab:red', label='LG')
    ax.set_yscale('log')
    ax.set_xlabel('TOF hits', loc='right')
    ax.set_ylabel('events', loc='top')
    ax.set_title('TOF HG (readout) vs LG (trigger) hits', loc='right')
    ax.text(0.5, 0.6, textbox, transform=fig.transFigure, fontsize=10)
    ax.legend(frameon=False, fontsize=8, ncol=3, bbox_to_anchor=(0.45,1.01),\
              bbox_transform=fig.transFigure)
    ax.set_xlim(left=-1, right=25) 
    fig_ratio = plt.figure(figsize=lo.FIGSIZE_A4_LANDSCAPE_HALF_HEIGHT)
    ax_ratio  = fig_ratio.gca()
    ax_ratio.set_xlabel('TOF hits', loc='right')
    ax_ratio.set_ylabel('ratio HG/LG', loc='top')
    ax_ratio.set_title('TOF HG (readout) / LG (trigger) hits', loc='right')
    ratio = dashi.histfuncs.histratio(h_nhits, h_nthits,\
                                  log=False, ylabel="HG/LG")
    ratio.scatter(color=markercolor, marker="o", markersize=3)
    ax_ratio.set_xlim(left=-1, right=25) 
    return (fig, fig_ratio)
    
#h_nrblnk.line(color='tab:red', label='RB LINK ID')

#--------------------------------------------------------

def tof_projection_xy(paddle_occupancy = {}, 
                      event            = None,
                      cmap             = matplotlib.colormaps['hot'],
                      paddle_style     = {'edgecolor' : 'w', 'lw' : 0.4},
                      show_cbar        = True,
                      overlay_panels   = False,
                      indicate_empty   = 'gray',
                      umbrella_only    = False,
                      lognorm          = False):
    """
    Show the projection of all paddles which
    are facing in z-direction
    These are the whole Umbrella as well as 
    CBE TOP + Bottom.
    While this plot can show the occupancy of TOF paddles,
    it can also be 'hijacked' to just highlight certain
    paddles.

    
    # Keyword Arguments:
        paddle_occupancy : The number of events per paddle
        event            : Plot hits from TofEvent or TofEventSummary
        cmap             : Colormap - can be lambda function
                           to return color value based on 
                           'occupancy' numbker
        show_cbar        : Show the colorbar on the figure
        overlay_panels   : Only return one axes, have the TOF CBE bottom
                           and CBE TOP panels overlaid over the umbrella
                           (or under it)
        indicate_empty   : In case we are using this for paddle occupancy,
                           indicate empty paddles with the given color instead
                           using a value from the color map. If this behavior is 
                           not desired, set this to an empty string.
        umbrella_only    : Show only the umbrella, add the colorbar next to it
    """
    if overlay_panels:
        fig = plt.figure(figsize=lo.FIGSIZE_A4_SQUARE)
        axs = [fig.gca()]
    else:
        fig, axs        = plt.subplots(1, 3, figsize=(18, 5), gridspec_kw={'width_ratios': [1, 1, 1]})
    
    all_paddles     = _gc.db.TofPaddle.all() 
    umb_paddles     = [k for k in all_paddles if k.paddle_id > 60 and k.paddle_id < 109]
    cbe_top_paddles = [k for k in all_paddles if k.panel_id == 1] 
    cbe_bot_paddles = [k for k in all_paddles if k.panel_id == 2]
    xmin, xmax      = -100,100
    ymin, ymax      = -100,100
    zmin, zmax      = -25, 120
    title           = 'TOF UMB/CBE TOP/CBE BOT xy projection'
    if event is not None:
        ts = np.array([h.t0 for h in event.hits])
        cm_norm_pts = plt.Normalize(vmin=min(ts), vmax=max(ts))
    
    if paddle_occupancy:
        # FIXME - this is a kludge for the ICRC approved plot!
        umb_norm = matplotlib.colors.Normalize(vmin=min(paddle_occupancy.values()), vmax=max(paddle_occupancy.values()))
        if umbrella_only:
            umb_occu = {k : paddle_occupancy[k] for k in paddle_occupancy if 60 < k < 109}
            paddle_occupancy = umb_occu
            if lognorm:
                umb_norm = matplotlib.colors.LogNorm(vmin=min(umb_occu.values()), vmax=max(umb_occu.values()))
            else:
                umb_norm = matplotlib.colors.Normalize(vmin=min(umb_occu.values()), vmax=max(umb_occu.values()))
    for pdl in umb_paddles:
        if paddle_occupancy:
            max_occu = max(paddle_occupancy.values())
            #color = cmap(paddle_occupancy[pdl.paddle_id]/max_occu)
            color = cmap(umb_norm(paddle_occupancy[pdl.paddle_id]))
            if not paddle_occupancy[pdl.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[0].add_patch(pdl.draw_xy(fill=True, edgecolor=color, facecolor=color))
            if pdl.panel_id == 7:
                axs[0].text(pdl.global_pos_x_l0 - 4.5, 75, f'{pdl.paddle_id}', color='w', fontsize=6)
            if pdl.panel_id == 10 or pdl.panel_id == 8:
                axs[0].text(pdl.global_pos_x_l0 - 4.5, 160, f'{pdl.paddle_id}', color='w', fontsize=6)
            if pdl.panel_id == 11 or pdl.panel_id == 13:
                axs[0].text(pdl.global_pos_x_l0 - 4.5, -170, f'{pdl.paddle_id}', color='w', fontsize=6)
            if pdl.panel_id == 9 or pdl.panel_id == 12:
                axs[0].text(-80, pdl.global_pos_y_l0 - 4.5, f'{pdl.paddle_id}', color='w', fontsize=6)


        else:
            if event is not None:
                axs[0].add_patch(pdl.draw_xy(fill=False,\
                                             edgecolor=paddle_style['edgecolor'],
                                             lw=paddle_style['lw'],
                                             facecolor='tab:blue'))#, alpha=0.3))
            else:
                axs[0].add_patch(pdl.draw_xy(fill=True, edgecolor='k', facecolor='w'))
        axs[0].set_xlabel('x [cm]', loc='right')
        axs[0].set_ylabel('y [cm]', loc='top')#, rotation=90)
        axs[0].set_aspect('equal')
        axs[0].set_xlim(-200, 200)
        axs[0].set_ylim(-200, 200)
        if overlay_panels:
            if umbrella_only:
                # this is a bad kludge. Sorry. 
                axs[0].set_title('TOF umbrella trigger hit occupancy', loc='right')
            else:
                axs[0].set_title('XY projection', loc='right')
        else:
            axs[0].set_title('TOF umbrella', loc='right')
    if event is not None:
        #print (hit_pids)
        umb_pids = [umbp.paddle_id for umbp in umb_paddles]
        for h in event.hits:
            if h.paddle_id in umb_pids:
        #for j, pdl_hit in enumerate(hit_pids):
            #if pdl_hit in umb_pids:
                #print (f"adding scatter for {0.1*h.x}, {0.1*h.y}")
                axs[0].scatter([0.1*h.x], [0.1*h.y], alpha = 0.8 , marker='o', s=100*h.edep,
                               lw=1.5, edgecolor=paddle_style['edgecolor'], color=cmap(cm_norm_pts(h.t0)))

    if umbrella_only:
        umb_occu = {k : paddle_occupancy[k] for k in paddle_occupancy if 60 < k < 109}
        if paddle_occupancy and show_cbar:
            cbar_ax = fig.add_axes([0.85, 0.0, 0.1, 1.0])
            cbar_ax.set_axis_off()
            sm = cm.ScalarMappable(cmap=cmap, norm=umb_norm)
            sm.set_array([min(umb_occu.values()), max(umb_occu.values())])
            ax = plt.sca(cbar_ax)
            plt.colorbar(sm, ax=cbar_ax, label='triggers', shrink=10.0, aspect=20)
        if not overlay_panels:
            fig.suptitle(title, x=0.9)
        axs[0].spines['top'].set_visible(True)
        axs[0].spines['right'].set_visible(True)

        return fig, axs
        pass 

    axid = 1
    if overlay_panels:
        axid = 0
    for pdl in cbe_top_paddles:   
        if paddle_occupancy:
            color = cmap(paddle_occupancy[pdl.paddle_id])
            if not paddle_occupancy[pdl.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[axid].add_patch(pdl.draw_xy(fill=True, edgecolor=color, facecolor=color))
            axs[axid].text(pdl.global_pos_x_l0 - 4, 75, f'{pdl.paddle_id}', color='w', fontsize=9)
        else:
            if event is not None:
                axs[axid].add_patch(pdl.draw_xy(fill=False,\
                                             edgecolor=paddle_style['edgecolor'],
                                             lw=paddle_style['lw'],
                                             facecolor='tab:blue'))#, alpha=0.3))
            else:
                axs[axid].add_patch(pdl.draw_xy(fill=True, edgecolor='k', facecolor='w'))
        if not overlay_panels:
            axs[axid].set_xlabel('x [cm]', loc='right')
            axs[axid].set_ylabel('y [cm]', loc='top')#, rotation=90)
            axs[axid].set_xlim(-100, 100)
            axs[axid].set_ylim(-100, 100)
            axs[axid].set_aspect('equal')
            axs[axid].set_title('CBE TOP', loc='right')

    if event is not None:
        cbpdl_pids = [cbpdl.paddle_id for cbpdl in cbe_top_paddles]
        for h in event.hits:
            if h.paddle_id in cbpdl_pids:
                axs[axid].scatter([0.1*h.x], [0.1*h.y], alpha = 0.8 , marker='o', s=100*h.edep,
                                lw=1.5, edgecolor=paddle_style['edgecolor'], color=cmap(cm_norm_pts(h.t0)))

    axid = 2
    if overlay_panels:
        axid = 0
    for pdl in cbe_bot_paddles:
        if paddle_occupancy:
            color = cmap(paddle_occupancy[pdl.paddle_id])
            if not paddle_occupancy[pdl.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[axid].add_patch(pdl.draw_xy(fill=True, edgecolor=color, facecolor=color))
            axs[axid].text(pdl.global_pos_x_l0 - 4, 75, f'{pdl.paddle_id}', color='w', fontsize=9)
        
        else:
            if event is not None:
                axs[axid].add_patch(pdl.draw_xy(fill=False,\
                                             edgecolor=paddle_style['edgecolor'],
                                             lw=paddle_style['lw'],
                                             facecolor='tab:blue'))#, alpha=0.3))
            else:
                axs[axid].add_patch(pdl.draw_xy(fill=True, edgecolor='k', facecolor='w'))
        if not overlay_panels:
            axs[axid].set_xlabel('x [cm]', loc='right')
            axs[axid].set_ylabel('y [cm]', loc='top')#, rotation=270)
            axs[axid].set_xlim(-100, 100)
            axs[axid].set_ylim(-100, 100)
            axs[axid].set_aspect('equal')
            axs[axid].set_title('CBE BOT')
    if event is not None:
        cbpdl_pids = [cbpdl.paddle_id for cbpdl in cbe_bot_paddles]
        for h in event.hits:
            if h.paddle_id in cbpdl_pids:
                axs[axid].scatter([0.1*h.x], [0.1*h.y], alpha = 0.8 , marker='o', s=100*h.edep,
                               lw=1.5, edgecolor=paddle_style['edgecolor'], color=cmap(cm_norm_pts(h.t0)))
        
    axs[0].spines['top'].set_visible(True)
    axs[0].spines['right'].set_visible(True)
    if not overlay_panels:
        axs[1].spines['top'].set_visible(True)
        axs[2].spines['top'].set_visible(True)
        axs[1].spines['right'].set_visible(True)
        axs[2].spines['right'].set_visible(True)
    if paddle_occupancy and show_cbar:
        cbar_ax = fig.add_axes([0.9, 0.0, 0.05, 1.0])
        cbar_ax.set_axis_off()
        sm = cm.ScalarMappable(cmap=cmap, norm=matplotlib.colors.Normalize())
        sm.set_array([0, 1])
        ax = plt.sca(cbar_ax)
        plt.colorbar(sm, ax=cbar_ax, label='Relative occupancy')
    if not overlay_panels:
        fig.suptitle(title, x=0.9)
    return fig, axs

#--------------------------------------------------------

def unroll_cbe_sides(paddle_occupancy = {},
                     event            = None,
                     cmap             = matplotlib.colormaps['hot'],
                     paddle_style    = {'edgecolor' : 'w', 'lw' : 0.4},
                     show_cbar        = True,
                     indicate_empty   = 'gray'
                     ):
    """
    Project the sides of the cube on xz and yz as well 
    as add the 'edge' paddles.

    While this plot can show the occupancy of TOF paddles,
    it can also be 'hijacked' to just highlight certain
    paddles.
    
    # Keyword Arguments:
        paddle_occupancy : The number of events per paddle
        event            : Plot a tof event
        cmap             : Colormap - can be lambda function
                           to return color value based on 
                           'occupancy' numbker
        show_cbar        : Show the colorbar on the figure
        indicate_empty   : In case we are using this for paddle occupancy,
                           indicate empty paddles with the given color instead
                           using a value from the color map. If this behavior is 
                           not desired, set this to an empty string.
    """
    fig, axs  = plt.subplots(1, 4, sharey=True,figsize=(22, 5), gridspec_kw={'width_ratios': [1, 1, 1, 1]})
    all_paddles = _gc.db.TofPaddle.all() 
    # normal +X
    cbe_front = [ k for k in all_paddles if k.panel_id==3]
    # edge normal +X+Y
    ep_1      = [ k for k in all_paddles if k.paddle_id==57]
    # normal +Y
    cbe_sb    = [ k for k in all_paddles if k.panel_id==4]
    # endge normal -X+Y
    ep_2      = [ k for k in all_paddles if k.paddle_id==58]
    # normal -X 
    cbe_back  = [ k for k in all_paddles if k.panel_id==5]
    # edge normal -X-Y
    ep_3      = [ k for k in all_paddles if k.paddle_id==59]
    # normal -Y 
    cbe_bb    = [ k for k in all_paddles if k.panel_id==6]
    # edge normal +X-Y
    ep_4      = [ k for k in all_paddles if k.paddle_id==60]

    xmin, xmax = -110,110
    ymin, ymax = -110,110
    zmin, zmax = -25, 120
    title      = 'Relative occupancy, xy projection'
    
    if event is not None:
        # coordinates are in mm
        hit_pids = [h.paddle_id for h in event.hits]
        xs = 0.1*np.array([h.x for h in event.hits])
        ys = 0.1*np.array([h.y for h in event.hits])
        zs = 0.1*np.array([h.z for h in event.hits])
        ts = np.array([h.t0 for h in event.hits])
        en = np.array([h.edep for h in event.hits])
        cm_norm_pts = plt.Normalize(vmin=min(ts), vmax=max(ts))

    ep = ep_1[0]
    cbe_front.append(ep_1[0])
    for pdl in cbe_front:
        if paddle_occupancy:
            color = cmap(paddle_occupancy[pdl.paddle_id])
            if not paddle_occupancy[pdl.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[0].add_patch(pdl.draw_yz(fill=True, edgecolor=color, facecolor=color))
        else:
            if event is not None:
                axs[0].add_patch(pdl.draw_yz(fill=False, facecolor='tab:blue', **paddle_style))
            else:
                axs[0].add_patch(pdl.draw_yz(fill=True, edgecolor='k', facecolor='w'))
    

    ##for pdl in cbe_front:
    ##    if paddle_occupancy:
    ##        color = cmap(paddle_occupancy[pdl.paddle_id])
    ##        axs[0].add_patch(pdl.draw_yz(fill=True, edgecolor=color, facecolor=color))
    ##    else:
    ##        if event is not None:
    ##            axs[0].add_patch(pdl.draw_yz(fill=False, facecolor='tab:blue', **paddle_style))
    ##        else:
    ##            axs[0].add_patch(pdl.draw_yz(fill=True, edgecolor='k', facecolor='w'))
    
    if event is not None:
        #print (hit_pids)
        cbe_front_pids = [k.paddle_id for k in cbe_front]
        for h in event.hits:
        #for j, pdl_hit in enumerate(hit_pids):
            if h.paddle_id in cbe_front_pids:
                axs[0].scatter([0.1*h.y], [0.1*h.z], alpha = 0.8 , marker='o', s=100*h.edep,
                               lw=1.5, edgecolor='k', color=cmap(cm_norm_pts(h.t0)))
    axs[0].set_xlabel('y [cm]', loc='right')
    axs[0].set_ylabel('z [cm]', loc='top')#, rotation=90)
    axs[0].set_aspect('equal')
    axs[0].set_xlim(-80, 90)
    axs[0].set_ylim(-10, 120)
    axs[0].set_title('CBE +X', loc='right')
    
    # +Y side
    ep = ep_2[0]
    cbe_sb.append(ep_2[0])
    #if paddle_occupancy:
    #    color = cmap(paddle_occupancy[ep.paddle_id])
    #    if not paddle_occupancy[ep.paddle_id] and indicate_empty:
    #        color = indicate_empty
    #    axs[1].add_patch(ep.draw_xz(fill=True, edgecolor=color, facecolor=color))
    #else:
    #    if event is not None:
    #        axs[1].add_patch(ep.draw_xz(fill=False, facecolor='tab:blue', **paddle_style))
    #    else:
    #        axs[1].add_patch(ep.draw_xz(fill=True, edgecolor='k', facecolor='w'))

    for pdl in cbe_sb:
        if paddle_occupancy:
            color = cmap(paddle_occupancy[pdl.paddle_id])
            if not paddle_occupancy[pdl.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[1].add_patch(pdl.draw_xz(fill=True, edgecolor=color, facecolor=color))
        else:
            if event is not None:
                axs[1].add_patch(pdl.draw_xz(fill=False, facecolor='tab:blue', **paddle_style))
            else:
                axs[1].add_patch(pdl.draw_xz(fill=True, edgecolor='k', facecolor='w'))
   
    if event is not None:
        #print (hit_pids)
        for j, pdl_hit in enumerate(hit_pids):
            if pdl_hit in [umbp.paddle_id for umbp in cbe_sb]:
                axs[1].scatter([xs[j]], [ys[j]], alpha = 0.8 , marker='o', s=100*en[j],
                               lw=1.5, edgecolor='k', color=cmap(cm_norm_pts(ts[j])))
    axs[1].set_xlabel('x [cm]', loc='right')
    #axs[1].set_ylabel('z [cm]', loc='top')#, rotation=90)
    axs[1].set_aspect('equal')
    axs[1].set_xlim(-90, 80)
    axs[1].set_ylim(-10, 120)
    axs[1].set_title('CBE +Y', loc='right')
    axs[1].invert_xaxis()

    ep = ep_3[0]
    cbe_back.append(ep_3[0])
    #if paddle_occupancy:
    #    color = cmap(paddle_occupancy[ep.paddle_id])
    #    if not paddle_occupancy[ep.paddle_id] and indicate_empty:
    #        color = indicate_empty
    #    axs[2].add_patch(ep.draw_yz(fill=True, edgecolor=color, facecolor=color))
    #else:
    #    if event is not None:
    #        axs[2].add_patch(ep.draw_yz(fill=False, facecolor='tab:blue', **paddle_style))
    #    else:
    #        axs[2].add_patch(ep.draw_yz(fill=True, edgecolor='k', facecolor='w'))
    for pdl in cbe_back:
        if paddle_occupancy:
            color = cmap(paddle_occupancy[pdl.paddle_id])
            if not paddle_occupancy[pdl.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[2].add_patch(pdl.draw_yz(fill=True, edgecolor=color, facecolor=color))
        else:
            if event is not None:
                axs[2].add_patch(pdl.draw_yz(fill=False, facecolor='tab:blue', **paddle_style))
            else:
                axs[2].add_patch(pdl.draw_yz(fill=True, edgecolor='k', facecolor='w'))
    if event is not None:
        #print (hit_pids)
        for j, pdl_hit in enumerate(hit_pids):
            if pdl_hit in [umbp.paddle_id for umbp in cbe_back]:
                axs[2].scatter([ys[j]], [zs[j]], alpha = 0.8 , marker='o', s=100*en[j],
                               lw=1.5, edgecolor='k', color=cmap(cm_norm_pts(ts[j])))
    axs[2].set_xlabel('y [cm]', loc='right')
    #axs[2].set_ylabel('z [cm]', loc='top')#, rotation=90)
    axs[2].set_xlim(-90, 80)
    axs[2].set_ylim(-10, 120)
    axs[2].set_aspect('equal')
    axs[2].invert_xaxis()
    axs[2].set_title('CBE -X', loc='right')
    
    # -Y side
    ep = ep_4[0]
    cbe_bb.append(ep_4[0])
    #if paddle_occupancy:
    #    color = cmap(paddle_occupancy[ep.paddle_id])
    #    if not paddle_occupancy[ep.paddle_id] and indicate_empty:
    #        color = indicate_empty
    #    axs[3].add_patch(ep.draw_xz(fill=True, edgecolor=color, facecolor=color))
    #else:
    #    if event is not None:
    #        axs[3].add_patch(ep.draw_xz(fill=False, edgecolor='k', facecolor='tab:blue', alpha=0.3))
    #    else:
    #        axs[3].add_patch(ep.draw_xz(fill=True, edgecolor='k', facecolor='w'))
    for pdl in cbe_bb:
        if paddle_occupancy:
            color = cmap(paddle_occupancy[pdl.paddle_id])
            if not paddle_occupancy[pdl.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[3].add_patch(pdl.draw_xz(fill=True, edgecolor=color, facecolor=color))
        else:
            if event is not None:
                axs[3].add_patch(pdl.draw_xz(fill=False,facecolor='tab:blue', **paddle_style))
            else:
                axs[3].add_patch(pdl.draw_xz(fill=True, edgecolor='k', facecolor='w'))
    if event is not None:
        #print (hit_pids)
        for j, pdl_hit in enumerate(hit_pids):
            if pdl_hit in [umbp.paddle_id for umbp in cbe_bb]:
                axs[3].scatter([xs[j]], [zs[j]], alpha = 0.8 , marker='o', s=100*en[j],
                               lw=1.5, edgecolor='k', color=cmap(cm_norm_pts(ts[j])))

    axs[3].set_xlabel('x [cm]', loc='right')
    #axs[3].set_ylabel('z [cm]', loc='top')#, rotation=90)
    axs[3].set_aspect('equal')
    axs[3].set_xlim(-80, 90)
    axs[3].set_ylim(-10, 120)
    axs[3].set_title('CBE +Y', loc='right')
    #axs[3].invert_xaxis()

    axs[0].spines['top'].set_visible(True)
    axs[1].spines['top'].set_visible(True)
    axs[2].spines['top'].set_visible(True)
    axs[3].spines['top'].set_visible(True)
    axs[0].spines['right'].set_visible(True)
    axs[1].spines['right'].set_visible(True)
    axs[2].spines['right'].set_visible(True)
    axs[3].spines['right'].set_visible(True)
    
    plt.subplots_adjust(wspace=0)
    
    if paddle_occupancy and show_cbar:
        cbar_ax = fig.add_axes([0.9, 0.0, 0.05, 1.0])
        cbar_ax.set_axis_off()
        sm = cm.ScalarMappable(cmap=cmap, norm=matplotlib.colors.Normalize())
        sm.set_array([0, 1])
        ax = plt.sca(cbar_ax)
        plt.colorbar(sm, ax=cbar_ax, label='Relative occupancy')
        fig.suptitle(title, x=0.9)
    return fig, axs

#--------------------------------------------------------

def unroll_cor(paddle_occupancy = {},
               event            = None,
               cmap             = matplotlib.colormaps['hot'],
               paddle_style     = {'edgecolor' : 'w', 'lw' : 0.4},
               show_cbar        = True,
               indicate_empty   = 'gray'):
    """
    Project the cortina on xz and yz as well 
    as add the 'edge' paddles.

    While this plot can show the occupancy of TOF paddles,
    it can also be 'hijacked' to just highlight certain
    paddles.
    
    # Keyword Arguments:
        paddle_occupancy : The number of events per paddle
        cmap             : Colormap - can be lambda function
                           to return color value based on 
                           'occupancy' numbker
        show_cbar        : Show the colorbar on the figure
        indicate_empty   : In case we are using this for paddle occupancy,
                           indicate empty paddles with the given color instead
                           using a value from the color map. If this behavior is 
                           not desired, set this to an empty string.
    """
    fig, axs  = plt.subplots(1, 4, sharey=True, figsize=(22, 5), gridspec_kw={'width_ratios': [1, 1, 1, 1]})
    all_paddles = _gc.db.TofPaddle.all()
    # normal +X
    cor_front = [k for k in all_paddles if k.panel_id==14]
    # edge normal +X+Y
    ep_1      = [k for k in all_paddles if k.panel_id==18]
    # normal +Y
    cor_sb    = [k for k in all_paddles if k.panel_id==15]
    # endge normal -X+Y
    ep_2      = [k for k in all_paddles if k.panel_id==19]
    # normal -X 
    cor_back  = [k for k in all_paddles if k.panel_id==16] 
    # edge normal -X-Y
    ep_3      = [k for k in all_paddles if k.panel_id==20]
    # normal -Y 
    cor_bb    = [k for k in all_paddles if k.panel_id==17]
    # edge normal +X-Y
    ep_4      = [k for k in all_paddles if k.panel_id==21]

    xmin, xmax = -100,130
    ymin, ymax = -25,175 # these are the z-coordinates
    title      = 'Relative occupancy, xy projection'
    
    if event is not None:
        # coordinates are in mm
        event_hits = sorted(event.hits, key=lambda x : x.t0)
        hit_pids = [h.paddle_id for h in event_hits]
        xs = 0.1*np.array([h.x for h in event_hits])
        ys = 0.1*np.array([h.y for h in event_hits])
        zs = 0.1*np.array([h.z for h in event_hits])
        ts = np.array([h.t0 for h in event_hits])
        en = np.array([h.edep for h in event_hits])
        cm_norm_pts = plt.Normalize(vmin=min(ts), vmax=max(ts))
    
    for ep in ep_1:
        if paddle_occupancy:
            color = cmap(paddle_occupancy[ep.paddle_id])
            if not paddle_occupancy[ep.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[0].add_patch(ep.draw_yz(fill=True, edgecolor=color, facecolor=color))
        else:
            if event is not None:
                axs[0].add_patch(ep.draw_yz(fill=False, facecolor='tab:blue', **paddle_style))
            else:
                axs[0].add_patch(ep.draw_yz(fill=True, edgecolor='k', facecolor='w'))
    if event is not None:
        for j, pdl_hit in enumerate(hit_pids):
            if pdl_hit in [pdl_.paddle_id for pdl_ in ep_1]:
                axs[0].scatter([ys[j]], [zs[j]], alpha = 0.8 , marker='o', s=100*en[j],
                               lw=1.5, edgecolor='k', color=cmap(cm_norm_pts(ts[j])))

    for pdl in cor_front:
        if paddle_occupancy:
            color = cmap(paddle_occupancy[pdl.paddle_id])
            if not paddle_occupancy[pdl.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[0].add_patch(pdl.draw_yz(fill=True, edgecolor=color, facecolor=color))
        else:
            if event is not None:
                axs[0].add_patch(pdl.draw_yz(fill=False, facecolor='tab:blue', **paddle_style))
            else:
                axs[0].add_patch(pdl.draw_yz(fill=True, edgecolor='k', facecolor='w'))
    
    if event is not None:
        for j, pdl_hit in enumerate(hit_pids):
            if pdl_hit in [pdl_.paddle_id for pdl_ in cor_front]:
                axs[0].scatter([ys[j]], [zs[j]], alpha = 0.8 , marker='o', s=100*en[j],
                               lw=1.5, edgecolor='k', color=cmap(cm_norm_pts(ts[j])))

    axs[0].set_xlabel('y [cm]', loc='right')
    axs[0].set_ylabel('z [cm]', loc='top')#, rotation=90)
    #axs[0].set_aspect('equal')
    axs[0].set_xlim(xmin, xmax)
    axs[0].set_ylim(ymin, ymax)
    axs[0].set_title('COR +X', loc='right')
    
    # +Y side
    for ep in ep_2:
        if paddle_occupancy:
            color = cmap(paddle_occupancy[ep.paddle_id])
            if not paddle_occupancy[ep.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[1].add_patch(ep.draw_xz(fill=True, edgecolor=color, facecolor=color))
        else:
            if event is not None:
                axs[1].add_patch(ep.draw_xz(fill=False, facecolor='tab:blue', **paddle_style))
            else:
                axs[1].add_patch(ep.draw_xz(fill=True, edgecolor='k', facecolor='w'))
    if event is not None:
        for j, pdl_hit in enumerate(hit_pids):
            if pdl_hit in [pdl_.paddle_id for pdl_ in ep_2]:
                axs[1].scatter([xs[j]], [zs[j]], alpha = 0.8 , marker='o', s=100*en[j],
                                lw=1.5, edgecolor='k', color=cmap(cm_norm_pts(ts[j])))

    for pdl in cor_sb:
        if paddle_occupancy:
            color = cmap(paddle_occupancy[pdl.paddle_id])
            if not paddle_occupancy[pdl.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[1].add_patch(pdl.draw_xz(fill=True, edgecolor=color, facecolor=color))
        else:
            if event is not None:
                axs[1].add_patch(pdl.draw_xz(fill=False, facecolor='tab:blue', **paddle_style))
            else:
                axs[1].add_patch(pdl.draw_xz(fill=True, edgecolor='k', facecolor='w'))
    if event is not None:
        for j, pdl_hit in enumerate(hit_pids):
            if pdl_hit in [pdl_.paddle_id for pdl_ in cor_sb]:
                axs[1].scatter([zs[j]], [zs[j]], alpha = 0.8 , marker='o', s=100*en[j],
                               lw=1.5, edgecolor='k', color=cmap(cm_norm_pts(ts[j])))
    axs[1].set_xlabel('x [cm]', loc='right')
    #axs[1].set_ylabel('z [cm]', loc='top')#, rotation=90)
    #axs[1].set_aspect('equal')
    axs[1].set_xlim(-1*xmax, -1*xmin)
    axs[1].set_ylim(ymin, ymax)
    axs[1].set_title('COR +Y', loc='right')
    axs[1].invert_xaxis()

    for ep in ep_3:
        if paddle_occupancy:
            color = cmap(paddle_occupancy[ep.paddle_id])
            if not paddle_occupancy[ep.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[2].add_patch(ep.draw_yz(fill=True, edgecolor=color, facecolor=color))
        else:
            if event is not None:
                axs[2].add_patch(ep.draw_yz(fill=False, facecolor='tab:blue', **paddle_style))
            else:
                axs[2].add_patch(ep.draw_yz(fill=True, edgecolor='k', facecolor='w'))
    if event is not None:
        for j, pdl_hit in enumerate(hit_pids):
            if pdl_hit in [pdl_.paddle_id for pdl_ in ep_3]:
                axs[2].scatter([ys[j]], [zs[j]], alpha = 0.8 , marker='o', s=100*en[j],
                               lw=1.5, edgecolor='k', color=cmap(cm_norm_pts(ts[j])))
    
    for pdl in cor_back:
        if paddle_occupancy:
            color = cmap(paddle_occupancy[pdl.paddle_id])
            if not paddle_occupancy[pdl.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[2].add_patch(pdl.draw_yz(fill=True, edgecolor=color, facecolor=color))
        else:
            if event is not None:
                axs[2].add_patch(pdl.draw_yz(fill=False, facecolor='tab:blue', **paddle_style))
            else:
                axs[2].add_patch(pdl.draw_yz(fill=True, edgecolor='k', facecolor='w'))
    
    if event is not None:
        for j, pdl_hit in enumerate(hit_pids):
            if pdl_hit in [pdl_.paddle_id for pdl_ in cor_back]:
                axs[2].scatter([ys[j]], [zs[j]], alpha = 0.8 , marker='o', s=100*en[j],
                               lw=1.5, edgecolor='k', color=cmap(cm_norm_pts(ts[j])))
    axs[2].set_xlabel('y [cm]', loc='right')
    #axs[2].set_ylabel('z [cm]', loc='top')#, rotation=90)
    axs[2].set_xlim(-1*xmax, -1*xmin)
    axs[2].set_ylim(ymin, ymax)
    #axs[2].set_aspect('equal')
    axs[2].invert_xaxis()
    axs[2].set_title('COR -X', loc='right')
    
    # -Y side
    for ep in ep_4:
        if paddle_occupancy:
            color = cmap(paddle_occupancy[ep.paddle_id])
            if not paddle_occupancy[ep.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[3].add_patch(ep.draw_xz(fill=True, edgecolor=color, facecolor=color))
        else:
            if event is not None:
                axs[3].add_patch(ep.draw_xz(fill=False, facecolor='tab:blue', **paddle_style))
            else:
                axs[3].add_patch(ep.draw_xz(fill=True, edgecolor='k', facecolor='w'))
    if event is not None:
        for j, pdl_hit in enumerate(hit_pids):
            if pdl_hit in [pdl_.paddle_id for pdl_ in ep_4]:
                axs[3].scatter([xs[j]], [zs[j]], alpha = 0.8 , marker='o', s=100*en[j],
                               lw=1.5, edgecolor='k', color=cmap(cm_norm_pts(ts[j])))

    for pdl in cor_bb:
        if paddle_occupancy:
            color = cmap(paddle_occupancy[pdl.paddle_id])
            if not paddle_occupancy[pdl.paddle_id] and indicate_empty:
                color = indicate_empty
            axs[3].add_patch(pdl.draw_xz(fill=True, edgecolor=color, facecolor=color))
        else:
            if event is not None:
                axs[3].add_patch(pdl.draw_xz(fill=False, facecolor='tab:blue', **paddle_style))
            else:
                axs[3].add_patch(pdl.draw_xz(fill=True, edgecolor='k', facecolor='w'))
    
    if event is not None:
        for j, pdl_hit in enumerate(hit_pids):
            if pdl_hit in [pdl_.paddle_id for pdl_ in cor_bb]:
                axs[3].scatter([xs[j]], [zs[j]], alpha = 0.8 , marker='o', s=100*en[j],
                               lw=1.5, edgecolor='k', color=cmap(cm_norm_pts(ts[j])))
    
    axs[3].set_xlabel('x [cm]', loc='right')
    #axs[3].set_ylabel('z [cm]', loc='top')#, rotation=90)
    #axs[3].set_aspect('equal')
    axs[3].set_xlim(xmin, xmax)
    axs[3].set_ylim(ymin, ymax)
    axs[3].set_title('COR +Y', loc='right')
    #axs[3].invert_xaxis()

    axs[0].spines['top'].set_visible(True)
    axs[1].spines['top'].set_visible(True)
    axs[2].spines['top'].set_visible(True)
    axs[3].spines['top'].set_visible(True)
    axs[0].spines['right'].set_visible(True)
    axs[1].spines['right'].set_visible(True)
    axs[2].spines['right'].set_visible(True)
    axs[3].spines['right'].set_visible(True)
    
    plt.subplots_adjust(wspace=0)

    if paddle_occupancy and show_cbar:
        cbar_ax = fig.add_axes([0.9, 0.0, 0.05, 1.0])
        cbar_ax.set_axis_off()
        sm = cm.ScalarMappable(cmap=cmap, norm=matplotlib.colors.Normalize())
        sm.set_array([0, 1])
        ax = plt.sca(cbar_ax)
        plt.colorbar(sm, ax=cbar_ax, label='Relative occupancy')
        fig.suptitle(title, x=0.9)
    return fig, axs

def tof_2dproj(event            = None,
               cmap             = matplotlib.colormaps['seismic'],
               paddle_style     = {'edgecolor' : 'w', 'lw' : 0.4},
               show_cbar        = True,
               no_ax_no_ticks   = False,
               cs_is_energy     = False,
               cnorm_max        = None) -> list:
    """
    Plots the entire TOF system in 2d projection, that is all panels 
    overlaid on each other

    # Keyword Arguments:
        event          : A TofEvent. If not None, then the hits will be shown
                         on top of the 2d projections 
        cs_is_energy   : Use the colorscale for energy instead of timing
        no_ax_no_ticks : Don't show any axis or axis ticks for a plain view 

    # Returns:
        list of figures, xy, xz, xy projections
    """
    projection_figures = []

    fig = plt.figure(figsize=lo.FIGSIZE_A4_SQUARE)
    ax = fig.gca()
    paddles         = _gc.db.TofPaddle.all()
    title           = 'XY projection'
    if event is not None:
        event.normalize_hit_times()
        ts = np.array([h.t0 for h in event.hits])
        if len(ts) > 0:
            cm_norm_pts = plt.Normalize(vmin=min(ts), vmax=max(ts))
            if cs_is_energy:
                edeps = np.array([h.edep for h in event.hits])
                if cnorm_max is None:
                    cm_norm_pts = plt.Normalize(vmin=min(edeps), vmax=max(edeps))
                else:
                    cm_norm_pts = plt.Normalize(vmin=0, vmax=cnorm_max)
    for pdl in paddles:
        if event is not None:
            ax.add_patch(pdl.draw_xy(fill=False,\
                                     edgecolor=paddle_style['edgecolor'],
                                     lw=paddle_style['lw'],
                                     facecolor='tab:blue'))#, alpha=0.3))
        else:
            ax.add_patch(pdl.draw_xy(fill=True, edgecolor='k', facecolor='w'))
    if event is not None:
        for h in event.hits:
            if cs_is_energy:
                ax.scatter([0.1*h.x], [0.1*h.y], alpha = 0.8 , marker='o', s=100*h.edep,
                           lw=1.5, edgecolor=paddle_style['edgecolor'], color=cmap(cm_norm_pts(h.edep)))
            else:
                ax.scatter([0.1*h.x], [0.1*h.y], alpha = 0.8 , marker='o', s=100*h.edep,
                           lw=1.5, edgecolor=paddle_style['edgecolor'], color=cmap(cm_norm_pts(h.t0)))
    ax.grid(0) 
    ax.set_xlabel('x [cm]', loc='right')
    ax.set_ylabel('y [cm]', loc='top')#, rotation=90)
    ax.set_aspect('equal')
    ax.set_xlim(-200, 200)
    ax.set_ylim(-200, 200)
    ax.set_title(title, loc='right')
    if no_ax_no_ticks:
        ax.set_axis_off()
    else:
        ax.spines['top'].set_visible(True)
        ax.spines['right'].set_visible(True)
    projection_figures.append(fig)
    
    # XZ projection
    fig = plt.figure(figsize=lo.FIGSIZE_A4_SQUARE)
    ax = fig.gca()
    title           = 'XZ projection'
    if event is not None:
        ts = np.array([h.t0 for h in event.hits])
        if len(ts) > 0:
            cm_norm_pts = plt.Normalize(vmin=min(ts), vmax=max(ts))

    for pdl in paddles:
        if event is not None:
            ax.add_patch(pdl.draw_xz(fill=False,\
                                     edgecolor=paddle_style['edgecolor'],
                                     lw=paddle_style['lw'],
                                     facecolor='tab:blue'))#, alpha=0.3))
        else:
            ax.add_patch(pdl.draw_xz(fill=True, edgecolor='k', facecolor='w'))
    if event is not None:
        for h in event.hits:
            if cs_is_energy:
                ax.scatter([0.1*h.x], [0.1*h.z], alpha = 0.8 , marker='o', s=100*h.edep,
                        lw=1.5, edgecolor=paddle_style['edgecolor'], color=cmap(cm_norm_pts(h.edep)))
            else:
                ax.scatter([0.1*h.x], [0.1*h.z], alpha = 0.8 , marker='o', s=100*h.edep,
                        lw=1.5, edgecolor=paddle_style['edgecolor'], color=cmap(cm_norm_pts(h.t0)))
    ax.grid(0) 
    ax.set_xlabel('x [cm]', loc='right')
    ax.set_ylabel('z [cm]', loc='top')#, rotation=90)
    ax.set_aspect('equal')
    ax.set_xlim(-200, 200)
    ax.set_ylim(-25, 250)
    ax.set_title(title, loc='right')

    if no_ax_no_ticks:
        ax.set_axis_off()
    else:
        ax.spines['top'].set_visible(True)
        ax.spines['right'].set_visible(True)
    if show_cbar:
        cbar_ax = fig.add_axes([0.9, 0.0, 0.05, 1.0])
        cbar_ax.set_axis_off()
        sm = cm.ScalarMappable(cmap=cmap, norm=matplotlib.colors.Normalize())
        #sm.set_array([0, 1])
        #ax = plt.sca(cbar_ax)
        if cs_is_energy:
            plt.colorbar(sm, ax=cbar_ax, label='Energy Dep. [MeV]')
        else:
            plt.colorbar(sm, ax=cbar_ax, label='Timing [ns]')
        fig.sca(ax)
    projection_figures.append(fig)

    # YZ projection
    fig = plt.figure(figsize=lo.FIGSIZE_A4_SQUARE)
    ax  = fig.gca()
    title           = 'XY projection'
    if event is not None:
        ts = np.array([h.t0 for h in event.hits])
        if len(ts) > 0:
            cm_norm_pts = plt.Normalize(vmin=min(ts), vmax=max(ts))

    for pdl in paddles:
        if event is not None:
            ax.add_patch(pdl.draw_yz(fill=False,\
                                     edgecolor=paddle_style['edgecolor'],
                                     lw=paddle_style['lw'],
                                     facecolor='tab:blue'))#, alpha=0.3))
        else:
            ax.add_patch(pdl.draw_yz(fill=True, edgecolor='k', facecolor='w'))
    if event is not None:
        for h in event.hits:
            if cs_is_energy:
                ax.scatter([0.1*h.y], [0.1*h.z], alpha = 0.8 , marker='o', s=100*h.edep,
                           lw=1.5, edgecolor=paddle_style['edgecolor'], color=cmap(cm_norm_pts(h.edep)))
            else:
                ax.scatter([0.1*h.y], [0.1*h.z], alpha = 0.8 , marker='o', s=100*h.edep,
                        lw=1.5, edgecolor=paddle_style['edgecolor'], color=cmap(cm_norm_pts(h.t0)))
    ax.grid(0) 
    ax.set_xlabel('y [cm]', loc='right')
    ax.set_ylabel('z [cm]', loc='top')#, rotation=90)
    ax.set_aspect('equal')
    ax.set_xlim(-200, 200)
    ax.set_ylim(-25, 250)
    ax.set_title(title, loc='right')
    if no_ax_no_ticks:
        ax.set_axis_off()
    else:
        ax.spines['top'].set_visible(True)
        ax.spines['right'].set_visible(True)
    projection_figures.append(fig)
    return projection_figures

#--------------------------------------------------------

def tof_hits_time_evolution(ev, line_color='k', t_err=0.35) -> plt.Figure: #, twindows=None):
    """
    A simple plot plotting normalized event
    times on the x axis and the energy deposition
    for each hit on the y axis.

    # Return:
        
    """
    fig = plt.figure(figsize=lo.FIGSIZE_A4_LANDSCAPE_HALF_HEIGHT)
    ax  = fig.gca()
    ev.normalize_hit_times()
    hits = sorted([h for h in ev.hits], key=lambda x : x.event_t0)
    if len(hits) == 0:
        return fig
    # the first hit
    first_hit = hits[0]
    ax.vlines(first_hit.event_t0,0, first_hit.edep, color=line_color)
    ax.fill_betweenx([0,first_hit.edep], first_hit.event_t0 - t_err, first_hit.event_t0 + t_err, color=line_color, alpha=0.2)
    prior_hit = first_hit
    for h in hits[1:]:
        # indicate lightspeed cleaning
        min_tdiff_cvac = 1e9*1e-3*prior_hit.distance(h)/299792458.0
        twindow        = prior_hit.event_t0 - t_err + min_tdiff_cvac;
        # lenient strategy
        ax.fill_betweenx([0,prior_hit.edep], prior_hit.event_t0, twindow, color='tab:red', alpha=0.4)
        # with an aggressive strategy, we could clean even more
        ax.fill_betweenx([0,prior_hit.edep], twindow, twindow + 2*t_err,
                         color='tab:blue',
                         alpha=0.2)
        ax.vlines(h.event_t0, 0, h.edep, color=line_color)
        ax.fill_betweenx([0, h.edep], h.event_t0 - t_err, h.event_t0 + t_err, color=line_color, alpha=0.2)
        prior_hit = h
    ax.set_xlabel('Event t0 [ns]', loc='right')
    ax.set_ylabel('Hit EDep', loc='top')
    ax.set_title('TOF Hitseries', loc='right')
    ax.set_ylim(bottom=0)
    cb.visual.adjust_minor_ticks(ax, which='both')
    return fig

#---------------------------------------------------------------------------------

def plot_waveforms(tof_ev, calib : dict = None, with_hits = False, skip_bins=0):
    """
    Return a list of figures with all the waveforms from a 
    specific tof event
    
    # Arguments:
        tof_ev    : TofEvent with waveforms
        calib     : A dictionary with RB calibrations
        with_hits : Indicate extracted hit time in the plots
        skip_bins : Zero the first [skip_bins]. This might be a 
                    helpful option in case there is a big spike 
                    in the beginning
    """
    wfs = tof_ev.waveforms
    figures, axes = [],[]
    hits = dict()
    for h in tof_ev.hits:
        hits[h.paddle_id] = h

    for wf in wfs:
        fig = plt.figure(figsize=lo.FIGSIZE_A4_LANDSCAPE_HALF_HEIGHT)
        ax  = fig.gca()
        if calib is None:
            adc_a = wf.adc_a
            adc_b = wf.adc_b
            if skip_bins:
                for k in range(skip_bins):
                    adc_a[k] = 0
                    adc_b[k] = 0
            ax.plot(adc_a, color='tab:blue', lw=1.2, label=f'{wf.paddle_id} A')
            ax.plot(adc_b, color='tab:red', lw=1.2, label=f'{wf.paddle_id} B')
            ax.set_xlabel('bin', loc='right')
            ax.set_ylabel('ADC', loc='top')
            ax.legend(frameon=False, loc='upper right')
        else:
            wf.calibrate(calib[wf.rb_id])
            voltages_a = wf.voltages_a
            voltages_b = wf.voltages_b
            if skip_bins:
                for k in range(skip_bins):
                    voltages_a[k] = 0
                    voltages_b[k] = 0
            ax.plot(wf.times_a, voltages_a, lw=0.9, color='tab:blue', label=f'{wf.paddle_id} A')
            ax.plot(wf.times_b, voltages_b, lw=0.9, color='tab:red', label=f'{wf.paddle_id} B')
            ax.set_xlabel('ns', loc='right')
            ax.set_ylabel('mV', loc='top')
            ax.legend(frameon=False, loc='upper right')
        if with_hits:
            try:
                ax.vlines(hits[wf.paddle_id].time_a, 0, max(voltages_a), lw=0.75, color='tab:blue')
            except KeyError:
                textbox = 'Hit extr. failed!'
                ax.text(0.2, 0.8, textbox, transform=fig.transFigure, fontsize=8)
                print ('No hit for waveform!')
            try:
                ax.vlines(hits[wf.paddle_id].time_b, 0, max(voltages_b), lw=0.75, color='tab:red')
            except KeyError:
                textbox = 'Hit extr. failed!'
                ax.text(0.2, 0.8, textbox, transform=fig.transFigure, fontsize=8)
                print ('No hit for waveform!')
        figures.append(fig)
        axes.append(ax)

    return figures, axes






"""
GAPS first guess & prototype event reconstructions

"""

from enum import Enum
import numpy as np

import iminuit
from iminuit.cost import LeastSquares as ls

import dashi as d
import json

#######################
#
## Errors from spatial dimensions of paddles/strips
## FIXME 
#ERR_X_TOF = 1
#ERR_Y_TOF = 1
#ERR_Z_TOF = 1
#
class FitStatus(Enum):
    Unknown        = 0
    DidNotConverge = 10
    Success        = 42

##############################################

def line3d_2pts(z, tu_x, tu_y, tu_z, tl_x, tl_y, tl_z):
    """
    describe line depending on z since that is our
    best constrained value. Version with 2 points 
    instead of direction vector, so we can constrain
    the point on the tof better
    """

    dx = tu_x - tl_x
    dy = tu_y - tl_y
    dz = tu_z - tl_z
    # a little cheating here
    if dz == 0: dz = 1e-3
    if dy == 0: dy = 1e-3
    if dx == 0: dx = 1e-3
    x  = tu_x + ((dx/dz) * (z - tu_z))
    y  = tu_y + ((dy/dz) * (z - tu_z))
    return x,y,z

##############################################

def line3d(z, x_a, y_a, z_a, dx, dy, dz):
    """
    describe line depending on z since that is our
    best constrained value

    This model has 6 free parameters, 3 for 
    the anchor point and 3 for the direction
    """
    # FIXME - proper treatment of ZeroDivisionError
    if dz == 0:
        dz = 1e-5
    x = x_a + ((dx/dz) * (z - z_a))
    y = y_a + ((dy/dz) * (z - z_a))
    return x,y,z

####################################################

class LeastSquares:
    """
    Generic least-squares cost function with error.
    """

    errordef = ls.errordef # for Minuit to compute errors correctly

    def __init__(self, model, x, y, z, x_err, y_err):
        self.model = model  # model predicts y for given x
        self.x     = x
        self.y     = y
        self.z     = z
        self.x_err = x_err
        self.y_err = y_err

    def __call__(self, *par):  # we accept a variable number of model parameters
        xm,ym,__ = self.model(self.z, *par)
        value    = np.sum(np.sqrt( ((self.x - xm ) ** 2 /self.x_err ** 2) + ((self.y - ym)**2 / self.y_err ** 2))) 
        #thesum = np.sum(((self.y - ym) ** 2 / self.y_err ** 2) + ((self.z - zm) ** 2 / self.z_err **2 ))
        #return thesum
        return value

#########################################################

def line_fit(xs, ys, zs, errs_x=None, errs_y=None, errs_z=None, search_anchor = False):
    """

    # Arguments:

        * search anchor : perform the fit multiple times, once per point. Select a different 
                          anchor point for the line each time (one of the hits) until all hits
                          have been used. Return the line with the best chi2.
    """

    assert len(xs) == len(ys) == len(zs)

    if errs_x is None:
        errs_x = 10*np.ones(len(xs)) 
    if errs_y is None:
        errs_y = 10*np.ones(len(ys))
    if errs_z is None:
        errs_z = 1*np.ones(len(zs))
    
    # rerutn a line for a convenience

    # the line3d takes z values and needs 6 parameters
    model = LeastSquares(line3d, xs, ys, zs, errs_x, errs_y)

    # start values
    if len(xs) < 2:
        print("Not enough points!")
        return
    x,y,z = xs[0],ys[0],zs[0]
    dx    = xs[1] - xs[0]
    dy    = ys[1] - ys[0]
    dz    = zs[1] - zs[0]
    m = iminuit.Minuit(model, x, y, z, dx, dy, dz)
    # force one of the points on the upper paddle
    # (and we know that this is 16 cm wide
    #dwidth = 0.635/2
    
    #m.limits = [(xs[0] - 8, xs[0] + 8), (-90,90), (zs[0]-dwidth, zs[0]+dwidth),\
    #             (None, None), (None, None), (None, None)]
    m.migrad()
    m.migrad()

    if search_anchor:
        #print ('RECO - iterative method (anchor search)')
        current_best = (m.fval, m.values)
        for k in range(1,len(xs)):
            x,y,z    = xs[k],ys[k],zs[k]
            dx,dy,dz = xs[k] - xs[k-1], ys[k] - ys[k-1], zs[k] - zs[k-1]
            m = iminuit.Minuit(model, x, y, z, dx, dy, dz)
            m.migrad()
            m.migrad()
            #recps.append((m.fval, m.values)
            if m.fval < current_best[0]:
                current_best = (m.fval, m.values)
            #print (f"RECO : Use anchor {x:.2f} {y:.2f} {z:.2f} ==> chi2 {m.fval:.2f}")
        def line(line_z_vals):
            return line3d(line_z_vals, *current_best[1])
        return line, current_best[0]  
    else:
        #print ('RECO - non-iterative method')
        def line(line_z_vals):
            return line3d(line_z_vals, *m.values)

        return line, m.fval

#########################################################3

class Reconstruction:

    def __init__(self, nbins = 100, active = False):
        self.active    = active
        self.CHIBINS   = np.linspace(0,500, nbins)
        self.COS2_BINS = np.linspace(0,1.1, nbins)
        self.BETA_BINS = np.linspace(-0.5,1.5, nbins)
        self.nevents   = 0
        self.event_cache_size = 1000000
        self.chi2_c    = []
        self.chi2      = d.histogram.hist1d(self.CHIBINS)
        self.cos2_c    = []
        self.cos2      = d.histogram.hist1d(self.COS2_BINS)
        self.beta_c    = []
        self.beta      = d.histogram.hist1d(self.BETA_BINS)
        self.finished  = False
        self.offsets = dict()
        offsets = json.load(open('offsets.json'))
        for k in offsets:
            k_int = int(k)
            # currently we only have intra-panel calibrations
            if k_int <= 12:
                self.offsets[k_int] = offsets[k]

    def __iadd__(self, other):
        self.chi2_c.extend(other.chi2_c)
        self.cos2_c.extend(other.cos2_c)
        self.beta_c.extend(other.beta_c)
        self.chi2 += other.chi2
        self.cos2 += other.cos2 
        self.beta += other.beta
        self.nevents += other.nevents
        return self

    def __add__(self, other):
        new_reco = Reconstruction()
        new_reco += self
        new_reco += other
        return new_reco

    def fill_histograms(self):
        if len(self.chi2_c) > self.event_cache_size:
            self.chi2.fill(np.array(self.chi2_c))
            self.chi2_c.clear()
            self.cos2.fill(np.array(self.cos2_c))
            self.cos2_c.clear()
            self.beta.fill(np.array(self.beta_c))
            self.beta_c.clear()

    def finish(self):
        ev_cache_size = self.event_cache_size
        self.event_cache_size = 1
        self.fill_histograms()
        self.finished = True
        self.event_cache_size = ev_cache_size


    def reco(self, ev):
        xs = [k[0] for k in ev.tracker_pointcloud]
        xs.extend([h.x for h in ev.tof.hits])

        ys = [k[1] for k in ev.tracker_pointcloud]
        ys.extend([h.y for h in ev.tof.hits])

        zs = [k[2] for k in ev.tracker_pointcloud]
        zs.extend([h.z for h in ev.tof.hits])

        ev.tof.normalize_hit_times()
        ev.tof.set_timing_offsets(self.offsets)
        ev.tof.remove_non_causal_hits()
        ev.tof.lightspeed_cleaning(0.35)
        ts = [(h.event_t0,h.x,h.y,h.z) for h in ev.tof.hits]
        ts = sorted(ts, key=lambda x: x[0])
        first_t = ts[0][0]
        last_t  = ts[-1][0]
        diff_h = last_t - first_t
        xs = np.array(xs)
        ys = np.array(ys)
        zs = np.array(zs)
        if len(xs) < 2:
            return
        reco = line_fit(xs, ys, zs)
        self.chi2_c.append(reco[1])
        
        # get the cosine by just selecting
        # two points on the line
        reco_f = reco[0]
        a = reco_f(0)
        b = reco_f(1000)
        d = np.sqrt((a[0] - b[0])**2 + (a[1] - b[1])**2 + (a[2] - b[2])**2)
        cos_theta = (b[2] - a[2])/d
        dist = np.sqrt((ts[-1][1] - ts[0][1])**2 + (ts[-1][2] - ts[0][2])**2 + (ts[-1][3] - ts[0][3])**2)
        beta = (dist/1000)/(diff_h*1e-9)/299792458  
        print (f'RECO BETA {beta}, DIST {dist}, TDIFF {diff_h}')
        self.beta_c.append(beta)
        self.cos2_c.append(cos_theta*cos_theta)
        return reco

#
#
#def linefit_trust_the_tof(tof_event, tracker_hits):
#    """
#    Try to fit a straight line, but "trust" the tof values.
#    That means we force the first hit on the upper and the
#    second hit on the lower tof. We use the tof hits as 
#    start values and constrain them on the actual paddles 
#    with the limits of the minimization.
#    
#    Keyword Arguments:
#        hits_blacklist (list) : allows to exclude hits from the 
#                                fit
#    """
#    xs         = np.array([h.x     for h in gaps_event.hits])
#    ys         = np.array([h.y     for h in gaps_event.hits])
#    zs         = np.array([h.z     for h in gaps_event.hits])
#    adc_data   = np.array([h.edep  for h in gaps_event.hits]) 
#    errs_x     = np.array([h.x_err for h in gaps_event.hits]) 
#    errs_y     = np.array([h.y_err for h in gaps_event.hits]) 
#
#
#    model    = LeastSquares(line3d_2pts, xs, ys, zs, errs_x, errs_y)
#    # hit 0 MUST be the upper and hit 1 MUST be the lower!!
#    # give the two tof hits as startvalues
#    m        = iminuit.Minuit(model,xs[0], ys[0], zs[0], xs[1], ys[1], zs[1])
#    #print (m.fixed)
#    # fix the z values
#    m.fixed['x2'] = True
#    m.fixed['x5'] = True
#    # force one of the points on the upper paddle
#    # (and we know that this is 16 cm wide
#    dwidth   = 0.635/2
#    m.limits = [(xs[0] - 8, xs[0] + 8), (-90,90), (zs[0]-dwidth, zs[0]+dwidth),\
#                (xs[1] - 8, xs[1] + 8), (-90,90), (zs[1]-dwidth, zs[1]+dwidth)]
#    
#    #print(m.limits)
#    m.migrad()
#    m.migrad()
#    m.hesse()
#    try:
#        #chi2 = m.fval/(len(xs) - m.nfit)
#        chi2 = m.fval
#    except ZeroDivisionError:
#        chi2 = np.nan
#    gaps_event.chi2 = chi2
#    # residuals
#    for jj,h in enumerate(gaps_event.hits):
#        reco_x, reco_y, reco_z = line3d_2pts(h.z, *m.values)
#        gaps_event.hits[jj].reco_residual_x = h.x - reco_x 
#        gaps_event.hits[jj].reco_residual_y = h.y - reco_y
#        gaps_event.hits[jj].reco_residual_z = h.z - reco_z 
#
#    def line(line_z_vals):
#        return line3d_2pts(line_z_vals, *m.values)
#
#    return line, gaps_event
#
#########################################################3
#
#
#def tracker_only_linefit(gaps_event):
#    """
#    This will fit a line only through the tracker hits. 
#    """
#
#    xs         = np.array([h.x     for h in gaps_event.hits if not (h.v_id.startswith('U') or h.v_id.startswith('L'))])
#    ys         = np.array([h.y     for h in gaps_event.hits if not (h.v_id.startswith('U') or h.v_id.startswith('L'))])
#    zs         = np.array([h.z     for h in gaps_event.hits if not (h.v_id.startswith('U') or h.v_id.startswith('L'))])
#    adc_data   = np.array([h.edep  for h in gaps_event.hits if not (h.v_id.startswith('U') or h.v_id.startswith('L'))]) 
#    errs_x     = np.array([h.x_err for h in gaps_event.hits if not (h.v_id.startswith('U') or h.v_id.startswith('L'))]) 
#    errs_y     = np.array([h.y_err for h in gaps_event.hits if not (h.v_id.startswith('U') or h.v_id.startswith('L'))]) 
#
#    model    = LeastSquares(line3d_2pts, xs, ys, zs, errs_x, errs_y)
#
#    # we use the first two hits (randomly) as the seed for the fit and
#    # constrain them within the limits of the tracker
#    m        = iminuit.Minuit(model,xs[0], ys[0], zs[0], xs[1], ys[1], zs[1])
#
#    dwidth   = 0.125
#    m.limits = [(xs[0] - 0.5, xs[0] + 0.5), (ys[0]-8, ys[0] + 8), (zs[0]-dwidth, zs[0]+dwidth),\
#                (xs[1] - 0.5, xs[1] + 0.5), (ys[1]-8, ys[1] + 8), (zs[1]-dwidth, zs[1]+dwidth)]
#
#    m.migrad()
#    m.migrad()
#    m.hesse()
#    try:
#        #chi2 = m.fval/(len(xs) - m.nfit)
#        chi2 = m.fval
#    except ZeroDivisionError:
#        chi2 = np.nan
#    gaps_event.chi2 = chi2
#    # residuals
#    for jj,h in enumerate(gaps_event.hits):
#        reco_x, reco_y, reco_z = line3d_2pts(h.z, *m.values)
#        gaps_event.hits[jj].reco_residual_x = h.x - reco_x 
#        gaps_event.hits[jj].reco_residual_y = h.y - reco_y
#        gaps_event.hits[jj].reco_residual_z = h.z - reco_z 
#
#    def line(line_z_vals):
#        return line3d_2pts(line_z_vals, *m.values)
#
#    return line, gaps_event
#
#
#class LineFit:
#    """
#    A simple line fit between 2 points
#    """
#
#    def __init__(self):
#        self.chi2_last_event = None
#        self.last_anchor     = None
#        self.last_direction  = None
#        self.last_fitstatus  = FitStatus.Unknown
#
#
#    def add_event(self, tof_ev, tracker_hits = [])
#        """
#        """
#        pass
#
#    def fit(self):
#        """
#        """
#        pass
